(defparameter *height* 20)
(defparameter *width* 50)
(defparameter *snake* nil)
(defparameter *nps* nil)
(defparameter *apple* nil)
(defparameter *dead-snakes* 0)
(defparameter *dead* '((-1 . -1)))
(defparameter *gameover* nil)

(defun memberp (x lst);xはletの要素？
  (if (null lst)
      nil
      (if (equal x (car lst))
	  t
	  (memberp x (cdr lst)))))

(defun npsp (p nps);p(場所、(x . y))にはnps(プレイヤー以外のヘビ)がいる？
  (memberp t (mapcar (lambda (snake)
		       (memberp p snake))
		     nps)))

(defun randput ();ランダム配置（何かあるところには置かない）
  (let ((p (cons (random *width*) (random *height*))))
    (if (or (memberp p *snake*)
	    (npsp p *nps*))
	(randput)
	p)))

(defun putnps (n);npsの配置
  (loop for i below n
	do (push (randput) *nps*)))

(defun move (snake x y p);文字通り、移動
  (let* ((sx (+ (caar snake) x))
	 (sy (+ (cdar snake) y))
	 (np (cons (mod sx *width*)
		   (mod sy *height*))))
    (cond
     ((equal np *apple*)
      (push np snake)
      (setf *apple* (randput)))
     ((memberp np *snake*)
      (princ "あなたは食べられた")
      (fresh-line)
      (setf *gameover* t))
     ((npsp np *nps*)
      (if p
	  (progn
	    (princ "あなたはヘビを食べてしまった")
	    (fresh-line)
	    (setf *gameover* t))
	(progn
	  (setf snake *dead*)
	  (incf *dead-snakes*))))
     (t (push np snake)
	(setf snake (cut-tail snake)))))
  snake)

(defun teleport (snake);テレポート、プレイヤーは使えない
  (push (randput) snake)
  (setf snake (cut-tail snake)))

(defun cut-tail (snake);リストの末尾を除いたリストを返す
  (labels ((collect-body (lst n)
			 (if (= n 0)
			     nil
			   (cons (car lst) (collect-body (cdr lst) (1- n))))))
	  (collect-body snake (1- (length snake)))))

(defun show-field ();フィールドの描画
  (fresh-line)
  (loop for y below *height*
     do (fresh-line)
       (progn (princ "|")
	      (loop for x below *width*
		 do (cond
		      ((memberp (cons x y) *snake*)
		       (if (equal (cons x y) (car *snake*))
			   (princ #\@)
			 (princ #\*)))
		      ((equal (cons x y) *apple*)
		       (princ #\a))
		      ((memberp t (mapcar (lambda (snake)
					    (equal (car snake) (cons x y)))
					  *nps*))
		       (princ #\o))
		      ((npsp (cons x y) *nps*)
		       (princ #\+))
		      (t (princ #\space))))
	      (princ "|"))))

(defun snake-loop ();ゲームループ
  (show-field)
  (fresh-line)
  (princ "[u]p [d]own [r]ight [l]eft:")
  (let ((str (read-line)))
    (if (equal str "quit")
	nil
      (progn
	(execmd (coerce str 'list))
	(if (null *gameover*)
	    (snake-loop)
	    (format t "Score : ~d" (length *snake*)))))))


(defun execmd (strlst);受け取ったコマンドの実行＋npsの行動の決定
  (if (null strlst)
      nil
    (progn
      (setf *nps* (mapcar (lambda (snake)
			    (let* ((sx (caar snake))
				   (sy (cdar snake))
				   (lx (- sx (car *apple*)))
				   (ly (- sy (cdr *apple*))))
			      (if (and (< 30 (+ (abs lx) (abs ly)))
				       (<= (length snake) (length *snake*))
				       (zerop (random 7)))
				  (teleport snake)
				  (let* ((dirs (sort (shuffle (list (cons '(1 . 0) (+ (abs (+ lx 1)) (abs ly)))
								    (cons '(-1 . 0) (+ (abs (- lx 1)) (abs ly)))
								    (cons '(0 . 1) (+ (abs lx) (abs (+ ly 1))))
								    (cons '(0 . -1) (+ (abs lx) (abs (- ly 1))))))
						     '<
						     :key #'cdr))
					 (dir (loop for l in dirs
						   when (not (memberp (cons (+ sx (caar l)) (+ sy (cdar l))) snake))
						   return (car l))))
				    (move snake (car dir) (cdr dir) nil)))))
			  *nps*))
      (if (< 0 *dead-snakes*)
	  (progn
	    (setf *nps* (del-key *dead* *nps*))
	    (loop for i below (* 2 *dead-snakes*)
		  do (push (list (randput)) *nps*))
	    (setf *dead-snakes* 0)))

      (cond
       ((equal (car strlst) #\u) (setf *snake* (move *snake* 0 -1 t)))
       ((equal (car strlst) #\d) (setf *snake* (move *snake* 0 1 t)))
       ((equal (car strlst) #\r) (setf *snake* (move *snake* 1 0 t)))
       ((equal (car strlst) #\l) (setf *snake* (move *snake* -1 0 t)))
       (t ()))
      (execmd (cdr strlst)))))

(defun start ();初期値を決めて、スタート
  (setf *snake* nil)
  (setf *nps* nil)
  (setf *apple* nil)
  (setf *dead-snakes* 0)
  (setf *gameover* nil)

  (setf *snake* (list (randput)))
  (setf *apple* (randput))
  (loop for i below 2
	do (push (list (randput)) *nps*))
  (snake-loop))

(defun shuffle (lst);リストの要素をランダムに並べ替える
  (if (null lst)
      nil
      (let ((el (nth (random (length lst)) lst)))
	(cons el (shuffle (del-key el lst))))))

(defun del-key (key lst);リストからkeyを除いたリストを返す
  (if (null lst)
      nil
      (if (equal key (car lst))
	  (del-key key (cdr lst))
	  (cons (car lst) (del-key key (cdr lst))))))
